-----------------------------------------------------------------------------
-- Graphical User Interface:
-----------------------------------------------------------------------------
local arcball = Arcball:create()
local Local = {}
Local.fovy = 29
Local.framerate = 30
Local.frame_count = 1
Local.tick_count = 1
Local.simulating = nil

GUI = {}

function GUI.checkReady()
    if GUI.teamOK and GUI.arenaOK then
        GUI.itemStart.active = IUP_YES
    else
        GUI.itemStart.active = IUP_NO
    end
    if GUI.arenaOK then GUI.itemChangeArena.active = IUP_YES end
end

function GUI.start()
        Simulator.restart()
        Local.startTime = gl_time()
        GUI.itemStart.active = IUP_NO
        Local.simulating = 1
        IupSetIdle(GUI.idle)
end

function GUI.showError(title, error)
    local body = {}
    gsub(error .. "\n", "([^\n]+)\n", function (line)
        tinsert(body, iuplabel{title = line})
    end)
    local button = iupbutton{
        title = "   OK   ",
        action = function(s) IupDestroy(s.owner) end
    }
    tinsert(body, iuphbox{iupfill{}, button, iupfill{}})
    local dialog = iupdialog{
        iupvbox{
            iupfill{size = 5},
            iuphbox{
                iupfill{size = 5},
                iupvbox(body),
                iupfill{size = 5}
            },
            iupfill{size = 5}
        };
    title = title,
    defaultesc = button
    }
    button.owner = dialog
    dialog:show()
end

function GUI.cleanUp()
    local last = gl_genlists(1)
    write("Cleaning up...")
    gl_deletelists(0, last+1)
    write("Done\n")
end

function GUI.setFrustum(fovy, aspect, zNear, zFar)
    local angle = (180 - fovy)/2
    local y = zNear/tan(angle)
    local x = y*aspect
    gl_matrixmode(GL_PROJECTION)
    gl_loadidentity()
    gl_frustum(-y, y, -x, x, zNear, zFar)
    gl_matrixmode(GL_MODELVIEW)
end

function GUI.changeProjection(w, h, ortho)
    w = w or Local.w
    h = h or Local.h
    Local.w = w
    Local.h = h
    local aspect = h/w
    if w <= h then arcball:set(w/2, h/2, w/2)
    else arcball:set(w/2, h/2, h/2) end
    IupGLMakeCurrent(GUI.glcMain)
    gl_viewport(0, 0, w, h)
    gl_matrixmode(GL_PROJECTION)
    gl_loadidentity()
    if ortho then
        if w > h then gl_ortho(-1.2*w/h, 1.2*w/h, -1.2, 1.2, -1, 1)
        else gl_ortho(-1.2, 1.2, -1.2*h/w, 1.2*h/w, -1, 1) end
    else
        GUI.setFrustum(Local.fovy, aspect, 4, 20)
    end
    gl_matrixmode(GL_MODELVIEW)
end

function GUI.idle()
    local elapsed = gl_time() - Local.startTime
    if elapsed / Local.tick_count > 0.02 then 
        Simulator.tick()
        Local.tick_count = Local.tick_count + 1
    else
        GUI.glcMain:action()
        Local.frame_count = Local.frame_count + 1
    end
    if Local.tick_count > 50 * Simulator.timeLimit then
        Simulator.showScores()
        IupSetIdle(nil)
        Local.simulating = nil
	return IUP_CLOSE
    end
    --Checks for Dead Robots:
    local dead = tremove(Simulator.DeadBots)
    if dead then
        GUI.showError("Error no rob ".. dead.name ..
            " from " .. dead.team, dead.err)
    end 
end

local gl_time = gl_time -- visibilitu issues
function GUI.timeElapsed()
    return gl_time() - Local.startTime
end

-----------------------------------------------------------------------------
-- GL Main Canvas:
-----------------------------------------------------------------------------
GUI.glcMain = iupglcanvas{
    rastersize="640x480",
    buffer = "DOUBLE",
    expand = "YES",
    depth_size="16"
}

-----------------------------------------------------------------------------
-- GL Main Canvas Context Menu:
-----------------------------------------------------------------------------
GUI.itemTitle = iupitem { title = "Modo 3D", active = IUP_NO }
GUI.itemStart = iupitem{
    title = "Iniciar",
    key = "I",
    active = IUP_NO,
    action = GUI.start
}
GUI.itemChangeArena = iupitem{
    title="Mudar Arena...",
    key="M",
    active = IUP_NO,
    action=function(self)
        local value, ret = IupGetFile("*.arn")
        if ret == 0 then 
            IupGLMakeCurrent(GUI.glcMain)
            local ret, err = Arena.load(value)
            if ret then 
                GUI.arenaOK = 1 
                GUI.arenafile = value
                GUI.itemShowMap.active = IUP_YES
            else IupMessage("Erro:", err) end
            GUI.checkReady()
        end
    end
}
GUI.itemSetup = iupitem{
        title="Configurao", 
        key="C", 
        action="IupShow(GUI.dlgSetup)"
}

GUI.itemRes = iupitem{
    title = "Modo 2D",
    key = "2",
    action = function()
        if GUI.schematic then
            gl_enable(GL_LIGHTING)
            GUI.changeProjection(Local.w, Local.h)
            GUI.schematic = nil
            GUI.itemRes.title = "Modo 2D"
            GUI.itemRes.key = "2"
            GUI.itemTitle.title = "Modo 3D"
        else
            gl_disable(GL_LIGHTING)
            GUI.changeProjection(Local.w, Local.h, 1)
            GUI.schematic = 1
            GUI.itemRes.title = "Modo 3D"
            GUI.itemRes.key = "3"
            GUI.itemTitle.title = "Modo 2D"
        end
    end
}

GUI.itemStep = iupitem{
    title = "Posio Inicial Aleatria",
    key = "A",
    action = function()
        if GUI.randompos then
            Simulator.toggleSlotFunction()
            GUI.randompos = nil
            GUI.itemStep.title = "Posio Inicial Aleatria"
            GUI.itemStep.key = "A"
        else
            Simulator.toggleSlotFunction()
            GUI.randompos = 1
            GUI.itemStep.title = "Posio Inicial Fixa"
            GUI.itemStep.key = "F"
        end
    end
}

GUI.itemShowMap = iupitem{
    title = "Mostrar Mapa de Presso",
    key = "P",
    active = IUP_NO,
    action = "GUI.dlgShowMap:show() GUI.glcShowMap:action()"
}

GUI.menuContext = iupmenu{
    GUI.itemTitle,
    iupseparator{},
    GUI.itemSetup,
    GUI.itemChangeArena,
    iupseparator{},
    GUI.itemStart,
    iupseparator{},
    GUI.itemRes,
    GUI.itemStep,
    --GUI.itemShowMap,
    iupitem{
        title="Sair", 
        key="r", 
        action = "return IUP_CLOSE"
    }
}
GUI.glcShowMap = iupglcanvas{
    rastersize="300x300",
    buffer = "DOUBLE",
    expand = "YES",
    depth_size="16"
}

function GUI.glcShowMap:action()
    IupGLMakeCurrent(self)
    gl_matrixmode(GL_PROJECTION)
    gl_loadidentity()
    gl_ortho(-1, 1, -1, 1, 1, -1)
    gl_matrixmode(GL_MODELVIEW)
    gl_clearcolor(0.2, 0.2, 0.2, 1)
    gl_clear()
    gl_loadidentity()
    local table = dofile(GUI.arenafile)
    local img = table.pmap
    local texture = gl_rgbtex(rgb_load(img))
    gl_disable(GL_LIGHTING)
    gl_enable(GL_COLOR_MATERIAL)
    gl_enable(GL_TEXTURE_2D)
    gl_texenv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)
    gl_texenv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, 1, 1, 1, 1)
    gl_bindtex(GL_TEXTURE_2D, texture)
    gl_color(1, 1, 1, 1)
    gl_material(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, 1, 1, 1, 1)
    gl_begin(GL_QUADS)
        gl_normal(0, 0, 1)
        gl_texcoord(0, 0)
        gl_vertex(-1, -1, 0)
        gl_texcoord(1, 0)
        gl_vertex(1, -1, 0)
        gl_texcoord(1, 1)
        gl_vertex(1, 1, 0)
        gl_texcoord(0, 1)
        gl_vertex(-1, 1, 0)
    gl_end()
    gl_disable(GL_TEXTURE_2D)
    IupGLSwapBuffers(self)
end

GUI.dlgShowMap = iupdialog{ 
    GUI.glcShowMap;
    title = "Mapa de Presso",
}

-----------------------------------------------------------------------------
-- Add Team Dialog
-----------------------------------------------------------------------------
GUI.buttonTeam1Color = iupbutton{
    title = "", 
    bgcolor="0 0 0", 
    size = "13x13",
    action = function(self)
        local convert = function(s)
            local r,g,b,_
            _, _, r, g, b = strfind(s, "(%d*) (%d*) (%d*)")
            if r and g and b then return r, g, b
            else return 0, 0, 0 end
        end
        local r, g, b = IupGetColor(IUP_CENTER, IUP_CENTER,
            convert(self.bgcolor))
        if r and g and b then
            GUI.buttonTeam1Color.bgcolor = format("%d %d %d", r, g, b)
        end
    end
}
GUI.buttonTeam2Color = iupbutton{
    title = "", 
    bgcolor="0 0 0", 
    size = "13x13",
    action = function(self)
        local convert = function(s)
            local r,g,b,_
            _, _, r, g, b = strfind(s, "(%d*) (%d*) (%d*)")
            if r and g and b then return r, g, b
            else return 0, 0, 0 end
        end
        local r, g, b = IupGetColor(IUP_CENTER, IUP_CENTER,
            convert(self.bgcolor))
        if r and g and b then
            GUI.buttonTeam2Color.bgcolor = format("%d %d %d", r, g, b)
        end
    end
}

GUI.textTeam1Name=iuptext{value="", size="80"}
GUI.textTeam2Name=iuptext{value="", size="80"}
GUI.textTeam1Script=iuptext{value="", size="100"}
GUI.textTeam2Script=iuptext{value="", size="100"}
GUI.textArena=iuptext{value="", size="100"}
GUI.dlgSetup = iupdialog{
    iuphbox{
        iupfill{size = 5},
        iupvbox{
            iupfill{size = 5},
            iupframe{title="Time 1:";
                iupvbox{
                    iupfill{size = 5},
                    iuphbox{
                        iupfill{size = 5},
                        iupvbox{
                            iuplabel{title = "Nome do Time:"},
                            GUI.textTeam1Name,
                        },
                        iupfill{},
                        GUI.buttonTeam1Color,
                        iupfill{size = 5};
                        alignment=IUP_ABOTTOM
                    },
                    iupfill{size = 5},
                    iuphbox{
                        iupfill{size = 5},
                        iuplabel{title = "Programa do Rob:"},
                    },
                    iuphbox{ 
                        iupfill{size = 5},
                        GUI.textTeam1Script,
                        iupfill{size = 5},
                        iupbutton{
                            title = "Procurar...",
                            size="45",
                            action = function(self)
                                local value, ret = IupGetFile("*.lua")
                                if ret == 0 then 
                                    GUI.textTeam1Script.value = value 
                                end
                            end
                        },
                        iupfill{size = 5}
                    },
                    iupfill{size = 5}
                }
            },
            iupfill{size = 10},
            iupframe{title="Time 2:";
                iupvbox{
                    iupfill{size = 5},
                    iuphbox{
                        iupfill{size = 5},
                        iupvbox{
                            iuplabel{title = "Nome do Time:"},
                            GUI.textTeam2Name,
                        },
                        iupfill{},
                        GUI.buttonTeam2Color,
                        iupfill{size = 5};
                        alignment=IUP_ABOTTOM
                    },
                    iupfill{size = 5},
                    iuphbox{
                        iupfill{size = 5},
                        iuplabel{title = "Programa do Rob:"},
                    },
                    iuphbox{ 
                        iupfill{size = 5},
                        GUI.textTeam2Script,
                        iupfill{size = 5},
                        iupbutton{
                            title = "Procurar...",
                            size="45",
                            action = function(self)
                                local value, ret = IupGetFile("*.lua")
                                if ret == 0 then 
                                    GUI.textTeam2Script.value = value 
                                end
                            end
                        },
                        iupfill{size = 5}
                    },
                    iupfill{size = 5}
                }
            },
            iupfill{size = 10},
            iupframe{title="Arena:";
                iupvbox{
                    iupfill{size=5},
                    iuphbox{
                        iupfill{size=5},
                        GUI.textArena,
                        iupfill{},
                        iupbutton{
                            title = "Procurar...",
                            size="45",
                            action=function(self)
                                local value, ret = IupGetFile("*.arn")
                                if ret == 0 then 
                                    GUI.textArena.value = value
                                end
                            end
                        },
                        iupfill{size=5}
                    },
                    iupfill{size=5}
                }
            },
            iupfill{size = 10},
            iuphbox{
                iupfill{},
                iupbutton{
                    title="OK",
                    size="40",
                    action=function(self)
                        local ret, err
                        ret, err = Simulator.addTeam(
                            GUI.textTeam1Name.value,
                            GUI.buttonTeam1Color.bgcolor,
                            2,
                            GUI.textTeam1Script.value
                        )
                        if not ret then 
                            IupMessage("Erro no time 1:", err)
                            return
                        end
                        ret, err = Simulator.addTeam(
                            GUI.textTeam2Name.value,
                            GUI.buttonTeam2Color.bgcolor,
                            2,
                            GUI.textTeam2Script.value
                        )
                        if not ret then 
                            IupMessage("Erro no time 2:", err)
                            Simulator.clean()
                            return
                        end
                        GUI.teamOK = 1
                        IupGLMakeCurrent(GUI.glcMain)
                        ret, err = Arena.load(GUI.textArena.value)
                        if ret then 
                            GUI.arenaOK = 1 
                            GUI.arenafile = value
                            GUI.itemShowMap.active = IUP_YES
                        else 
                            Simulator.clean()
                            IupMessage("Erro na Arena:", err) 
                        end
                        GUI.itemSetup.active = IUP_NO
                        GUI.checkReady()
                        GUI.dlgSetup:hide()
                    end
                },
                iupfill{},
                iupbutton{
                    title="Cancelar",
                    size="45",
                    action="GUI.dlgSetup:hide()"
                },
                iupfill{}
            },
            iupfill{size = 5}
        },
        iupfill{size = 5}
    };
    title = "Times",
    resize = "NO"
}

-----------------------------------------------------------------------------
-- GL Main Canvas Callbacks:
-----------------------------------------------------------------------------
function GUI.glcMain:action()
    IupGLMakeCurrent(self)
    gl_clearcolor(0.2, 0.2, 0.6, 1)
    gl_clear()
    gl_loadidentity()
    if GUI.schematic then
        Simulator.drawSchematic()
    else
        gl_light(GL_LIGHT1, GL_POSITION, 0, 0, 0, 1)
        gl_translate(0, 0, -6)
        gl_rotate(arcball:torotation())
        gl_light(GL_LIGHT0, GL_POSITION, -1, -1, 1, 1)
        Simulator.draw3D()
     end
    IupGLSwapBuffers(self)
end

function GUI.glcMain:resize(w, h)
    if GUI.schematic then GUI.changeProjection(w, h, 1)
    else GUI.changeProjection(w, h) end
end

function GUI.glcMain:button(b, s, x, y)
    local y = Local.h - y
    if b == IUP_BUTTON1 then
        if s == 1 then
            arcball:down(x, y)
            Local.drag = 1
        else
            arcball:up(x, y)
            Local.drag = nil
        end
    elseif b == IUP_BUTTON3 and s == 0 then
        IupPopup(GUI.menuContext, IUP_MOUSEPOS, IUP_MOUSEPOS)
    elseif b == IUP_BUTTON2 and s == 0 then
        arcball:reset()
        self:action()
    end
    return IUP_DEFAULT
end

function GUI.glcMain:motion(x, y)
    if Local.drag then
        local y = Local.h - y
        arcball:motion(x, y)
        if not Local.simulating then GUI.glcMain:action() end
    end
end

function GUI.glcMain:k_any(k)
    if k == K_UP then
        if not GUI.schematic then Local.fovy = Local.fovy - 0.5 end
    end
    if k == K_DOWN then
        if not GUI.schematic then Local.fovy = Local.fovy + 0.5 end
    end
    if k == K_Q or k == K_q then
        Simulator.showScores()
        GUI.cleanUp()
        IupSetIdle(nil)
        return IUP_CLOSE
    end
    if k == K_D or k == K_d then
        if GUI.detailed then GUI.detailed = nil
        else GUI.detailed = 1 end
    end
    local wxh = GUI.glcMain.rastersize
    local _, w, h
    _, _, w = strfind(wxh, "([^x]*)x")
    _, _, h = strfind(wxh, "x([^x]*)")
    GUI.changeProjection(w, h, GUI.schematic)
end
-----------------------------------------------------------------------------
-- Defines the main dialog:
-----------------------------------------------------------------------------
GUI.dlgArena = iupdialog{
    GUI.glcMain;
    title = "Simulao",
    close = "GUI.cleanUp(); IupSetIdle(nil); return IUP_CLOSE",
}
