---------------------------------------------------------------------------
--Scene Loader Module
---------------------------------------------------------------------------
--(Loads a scene exported from 3DS Max, into a calllist)

local Public, Private = {}, {}
Scene = Public

Private.eventtable = { index = {} }
Private.BLACK = {r = 0, g = 0, b = 0}

function Private.fit(scene)
    local factor
    if scene.width > scene.height then
        if scene.width > scene.depth then factor = 2/scene.width
        else factor = 2/scene.depth end
    else
        if scene.height > scene.depth then factor = 2/scene.height
        else factor = 2/scene.depth end
    end
    return factor
end

function Private.sortbymaterial(scene)
    sort(scene.object, function(o1, o2)
        return o1.material_ref < o2.material_ref
    end)
    for i = 1, getn(scene.object) do
        sort(scene.object[i].face, function(f1, f2)
            return f1.material_id < f2.material_id
        end)
    end
end

function Private.bbox(scene)
    local xmin = scene.object[1].vertex[1].x
    local xmax = scene.object[1].vertex[1].x
    local ymin = scene.object[1].vertex[1].y
    local ymax = scene.object[1].vertex[1].y
    local zmin = scene.object[1].vertex[1].z
    local zmax = scene.object[1].vertex[1].z
    for i = 1, getn(scene.object) do
        local object = scene.object[i]
        for j = 0, getn(object.vertex) do
            local vertex = object.vertex[j]
            xmin = min(xmin, vertex.x)
            ymin = min(ymin, vertex.y)
            zmin = min(zmin, vertex.z)
            xmax = max(xmax, vertex.x)
            ymax = max(ymax, vertex.y)
            zmax = max(zmax, vertex.z)
        end
    end
    return xmin, ymin, zmin, xmax, ymax, zmax
end

function Private.getMaterial(scene, object, face)
    local material_ref = object.material_ref
    local material_id = face.material_id
    local material = scene.material[material_ref]
    local standard
    if material[0] then
        local n = getn(material) + 1
        local id = mod(material_id, n)
        standard = material[id]
    else standard = material end
    return standard
end

function Private.setMaterial(material)
    local color
    if material.name ~= "team" then
        color = material.diffuse
        gl_material(GL_FRONT_AND_BACK, GL_DIFFUSE, color.r, color.g, color.b)
        color = material.ambient
        gl_material(GL_FRONT_AND_BACK, GL_AMBIENT, color.r, color.g, color.b)
    end
    color = material.specular
    gl_material(GL_FRONT_AND_BACK, GL_SPECULAR, color.r, color.g, color.b)
    gl_material(GL_FRONT_AND_BACK, GL_SHININESS, material.shininess)
    color = material.emission
    gl_material(GL_FRONT_AND_BACK, GL_EMISSION, color.r, color.g, color.b)
    if material.diffuse_tex then
        gl_enable(GL_TEXTURE_2D)
        gl_texenv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)
        gl_texenv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, 1, 1, 1, 1)
        gl_bindtex(GL_TEXTURE_2D, material.diffuse_tex)
    else 
        gl_disable(GL_TEXTURE_2D) 
    end
end

function Private.render(scene)
    local last_material = {name = "dummy"}
    for i = 1, getn(scene.object) do
        local object = scene.object[i]
        gl_begin(GL_TRIANGLES)
        for j = 0, getn(object.face) do
            local face = object.face[j]
            local material = Private.getMaterial(scene, object, face)
            if last_material ~= material then
                gl_end()
                if material.name == "team" then
                    gl_popattrib()
                elseif last_material.name == "team" then 
                    gl_pushattrib()
                end
                Private.setMaterial(material)
                gl_begin(GL_TRIANGLES)
                last_material = material
            end
            for k = 1, 3 do
                local id = face.vertex_id[k]
                local vertex = object.vertex[id]
                local normal = face.vertex_normal[id]
                gl_normal(normal.x, normal.y, normal.z)
                if face.texvertex_id then
                    id = face.texvertex_id[k]
                    local texvertex = object.texvertex[id]
                    gl_texcoord(texvertex.u, texvertex.v)
                end
                gl_vertex(vertex.x, vertex.y, vertex.z)
            end
        end
        gl_end()
    end
    if last_material.name == "team" then gl_pushattrib() end
    if last_material.diffuse_tex then gl_disable(GL_TEXTURE_2D) end
end

function Private.loadmaps(material)
    local diffuse_map = material.diffuse_map
    if diffuse_map then
        local map = rgb_load(diffuse_map.bitmap)
        if map then
            local diffuse_tex = gl_rgbtex(map)
            material.diffuse_tex = diffuse_tex
        end
    end
end

function Private.adjustmaterial(material)
    if material[0] then
        local i = 0
        while material[i] do
            Private.adjustmaterial(material[i])
            i = i + 1
        end
    else
        local glossiness = material.glossiness or 0
        material.shininess = glossiness * 128
        local level = material.specular_level or 0
        local specular = material.specular or Private.BLACK
        specular = { 
            r = specular.r * level,
            g = specular.g * level,
            b = specular.b * level,
        }
        material.specular = specular
        material.ambient = material.ambient or Private.BLACK
        material.diffuse = material.diffuse or Private.BLACK
        material.emission = material.emission or Private.BLACK
        material.name = strlower(material.name)
        Private.loadmaps(material)
    end
end

function Public.load(file)
    local scene = dofile(file)
    local xmin, ymin, zmin, xmax, ymax, zmax = Private.bbox(scene)
    local width = xmax - xmin
    local height = ymax - ymin
    local depth = zmax - zmin
    local xc = xmin + width/2
    local yc = ymin + height/2
    local zc = zmin + depth/2
    scene.xc = xc
    scene.yc = yc
    scene.zc = zc
    scene.width = width
    scene.height = height
    scene.depth = depth
    Private.sortbymaterial(scene)
    Private.adjustmaterial(scene.material)
    --scene.list = gl_genlists(1)
    --gl_newlist(scene.list, GL_COMPILE)
    --Private.render(scene)
    --gl_endlist()
    --scene.object = nil
    --scene.material = nil
    scene.factor = Private.fit(scene)
    eventtable(scene, Private.eventtable)
    return scene
end

function Private.eventtable.index.draw(scene)
    gl_pushattrib()
    --gl_calllist(scene.list)
    Private.render(scene) 
    gl_popattrib()
end
